"""
:mod:`coinaddr`
~~~~~~~~~~~

A crypto-currency address inspection/validation library for python.

Usage::

    >>> import coinaddr
    >>> coinaddr.validate('btc', b'1BoatSLRHtKNngkdXEeobR76b53LETtpyT')
    ValidationResult(name='bitcoin', ticker='btc',
    ...              address=b'1BoatSLRHtKNngkdXEeobR76b53LETtpyT', valid=True,
    ...              network='main')

:copyright: (c) 2018 by Joseph Black.
:license: MIT, see LICENSE for more details.
"""

__version__ = '1.0.1'

from . import interfaces, currency, validation
from .validation import validate
from .currency import Currency
from .validation import ValidatorBase, Base58CheckValidator, EthereumValidator



# pylint: disable=no-member

"""
:mod:`coinaddr.validation`
~~~~~~~~~~~~~~~~~~~~~~~~

Various validation machinery for validating cryptocurrency addresses.
"""

import re
from hashlib import sha256
import functools
import operator

from zope.interface import implementer, provider
import attr
import sha3
import base58check
import math
from binascii import unhexlify

from .interfaces import (
    INamedSubclassContainer, IValidator, IValidationRequest,
    IValidationResult, ICurrency
    )
from .base import NamedSubclassContainerBase
from . import currency


@provider(INamedSubclassContainer)
class Validators(metaclass=NamedSubclassContainerBase):
    """Container for all validators."""


class ValidatorMeta(type):
    """Register validator classes on Validators.validators."""

    def __new__(mcs, cls, bases, attrs):
        new = type.__new__(mcs, cls, bases, attrs)
        if new.name:
            Validators[new.name] = new
        return new


@attr.s(cmp=False, slots=True)
class ValidatorBase(metaclass=ValidatorMeta):
    """Validator Interface."""

    name = None

    request = attr.ib(
        type='ValidationRequest',
        validator=[
            lambda i, a, v: type(v).__name__ == 'ValidationRequest',
            attr.validators.provides(IValidationRequest)
            ]
    )

    def validate(self):
        """Validate the address type, return True if valid, else False."""

    @property
    def network(self):
        """Return the network derived from the network version bytes."""


@attr.s(frozen=True, slots=True, cmp=False)
@implementer(IValidator)
class Base58CheckValidator(ValidatorBase):
    """Validates Base58Check based cryptocurrency addresses."""

    name = 'Base58Check'

    def validate(self):
        """Validate the address."""
        if 25 > len(self.request.address) > 35:
            return False

        abytes = base58check.b58decode(
            self.request.address, **self.request.extras)

        if self.network == "":
            return False

        checksum = sha256(sha256(abytes[:-4]).digest()).digest()[:4]
        if abytes[-4:] != checksum:
            return False

        return self.request.address == base58check.b58encode(
            abytes, **self.request.extras)

    @property
    def network(self):
        """Return network derived from network version bytes."""
        abytes = base58check.b58decode(
            self.request.address, **self.request.extras)

        for name, networks in self.request.currency.networks.items():
            for netw in networks:
                if netw != 0:
                    prefixlen = math.ceil(math.floor((math.log(netw) / math.log(2)) + 1) / 8)
                else:
                    prefixlen = 1

                if abytes[:prefixlen] == unhexlify("%02x" % (netw)):
                    return name

        return ""

@attr.s(frozen=True, slots=True, cmp=False)
@implementer(IValidator)
class EthereumValidator(ValidatorBase):
    """Validates ethereum based crytocurrency addresses."""

    name = 'Ethereum'
    non_checksummed_patterns = (
        re.compile("^(0x)?[0-9a-f]{40}$"), re.compile("^(0x)?[0-9A-F]{40}$")
        )


    def validate(self):
        """Validate the address."""
        address = self.request.address.decode()
        if any(bool(pat.match(address))
               for pat in self.non_checksummed_patterns):
            return True
        addr = address[2:] if address.startswith('0x') else address
        addr_hash = sha3.keccak_256(addr.lower().encode('ascii')).hexdigest()
        for i, letter in enumerate(addr):
            if any([
                    int(addr_hash[i], 16) >= 8 and letter.upper() != letter,
                    int(addr_hash[i], 16) < 8 and letter.lower() != letter
            ]):
                return False
        return True


    #def validate(self):
    #    """Validate the address."""
    #    address = self.request.address.decode()
    #    if any(bool(pat.match(address))
    #           for pat in self.non_checksummed_patterns):
    #        return True
    #    addr = address.lstrip('0x')
    #    addr_hash = sha3.keccak_256(addr.lower().encode('ascii')).hexdigest()
    #    for i in range(0, len(addr)):
    #        if any([
    #                int(addr_hash[i], 16) > 7 and addr[i].upper() != addr[i],
    #                int(addr_hash[i], 16) <= 7 and addr[i].lower() != addr[i]
    #        ]):
    #            return False
    #    return True

    @property
    def network(self):
        """Return network derived from network version bytes."""
        return 'both'


@attr.s(frozen=True, slots=True, cmp=False)
@implementer(IValidationRequest)
class ValidationRequest:
    """Contain the data and helpers as an immutable request object."""

    currency = attr.ib(
        type=currency.Currency,
        converter=currency.Currencies.get,
        validator=[
            attr.validators.instance_of(currency.Currency),
            attr.validators.provides(ICurrency)
            ])
    address = attr.ib(
        type=bytes,
        converter=lambda a: a if isinstance(a, bytes) else a.encode('ascii'),
        validator=attr.validators.instance_of(bytes))

    @property
    def extras(self):
        """Extra arguments for passing to decoder, etc."""
        extras = dict()
        if self.currency.charset:
            extras.setdefault('charset', self.currency.charset)
        return extras

    @property
    def networks(self):
        """Concatenated list of all version bytes for currency."""
        networks = tuple(self.currency.networks.values())
        return functools.reduce(operator.concat, networks)

    def execute(self):
        """Execute this request and return the result."""
        validator = Validators.get(self.currency.validator)(self)
        return ValidationResult(
            name=self.currency.name,
            ticker=self.currency.ticker,
            address=self.address,
            valid=validator.validate(),
            network=validator.network
            )


@attr.s(frozen=True, slots=True, cmp=False)
@implementer(IValidationResult)
class ValidationResult:
    """Contains an immutable representation of the validation result."""

    name = attr.ib(
        type=str,
        validator=attr.validators.instance_of(str))
    ticker = attr.ib(
        type=str,
        validator=attr.validators.instance_of(str))
    address = attr.ib(
        type=bytes,
        validator=attr.validators.instance_of(bytes))
    valid = attr.ib(
        type=bool,
        validator=attr.validators.instance_of(bool))
    network = attr.ib(
        type=str,
        validator=attr.validators.instance_of(str))

    def __bool__(self):
        return self.valid


def validate(currency, address):
    """Validate the given address according to currency type.

    This is the main entrypoint for using this library.

    :param currency str: The name or ticker code of the cryptocurrency.
    :param address (bytes, str): The crytocurrency address to validate.
    :return: a populated ValidationResult object
    :rtype: :inst:`ValidationResult`

    Usage::

      >>> import coinaddr
      >>> coinaddr.validate('btc', b'1BoatSLRHtKNngkdXEeobR76b53LETtpyT')
      ValidationResult(name='bitcoin', ticker='btc',
      ...              address=b'1BoatSLRHtKNngkdXEeobR76b53LETtpyT',
      ...              valid=True, network='main')

    """
    request = ValidationRequest(currency, address)
    return request.execute()



import unittest

import coinaddr

from coinaddr.interfaces import (
    INamedSubclassContainer, INamedInstanceContainer, ICurrency, IValidator,
    IValidationRequest, IValidationResult
    )
from coinaddr.currency import Currencies, Currency
from coinaddr.validation import (
    Validators, ValidatorBase, ValidationRequest, ValidationResult,
    Base58CheckValidator, EthereumValidator
    )


TEST_DATA = [
    ('bitcoin', 'btc', b'1BoatSLRHtKNngkdXEeobR76b53LETtpyT', 'main'),
    ('bitcoin', 'btc', b'n2nzi7xDTrMVK9stGpbK3BtrpBCJfH7LRQ', 'test'),
    ('bitcoin', 'btc', b'3QJmV3qfvL9SuYo34YihAf3sRCW3qSinyC', 'main'),
    ('bitcoin-cash', 'bch', b'1BoatSLRHtKNngkdXEeobR76b53LETtpyT', 'main'),
    ('bitcoin-cash', 'bch', b'n2nzi7xDTrMVK9stGpbK3BtrpBCJfH7LRQ', 'test'),
    ('bitcoin-cash', 'bch', b'3QJmV3qfvL9SuYo34YihAf3sRCW3qSinyC', 'main'),
    ('litecoin', 'ltc', b'LeF6vC9k1qfFDEj6UGjM5e4fwHtiKsakTd', 'main'),
    ('litecoin', 'ltc', b'mkwV3DZkgYwKaXkphBtcXAjsYQEqZ8aB3x', 'test'),
    ('neocoin', 'neo', b'AL9fzczwjV6ynoFAJVz4fBDu4NYLG6MBwm', 'both'),
    ('dogecoin', 'doge', b'DAnBU2rLkUgQb1ZLBJd6Bm5pZ45RN4TQC4', 'main'),
    ('dogecoin', 'doge', b'njscgXBB3HUUTXH7njim1Uw82PF9da4R8k', 'test'),
    ('dashcoin', 'dash', b'XsVkhTxLjzdXP1xZWtEFRj1mDhWcU6d8tE', 'main'),
    ('dashcoin', 'dash', b'yPv7h2i8v3dJjfSH4L3x91JSJszjdbsJJA', 'test'),
    ('ether-zero', 'etz', b'900ff070d37657cdf8016bca0d60cb493ebf7f83', 'both'),
    ('ethereum-classic', 'etc',
     b'0x900ff070d37657cdf8016bca0d60cb493ebf7f83', 'both'),
    ('ethereum', 'eth', b'900Ff070D37657cdF8016BcA0D60CB493EBf7f83', 'both'),
    ('ethereum-classic', 'etc',
     b'0x900Ff070D37657cdF8016BcA0D60CB493EBf7f83', 'both')
]


class TestCoinaddr(unittest.TestCase):
    def test_validation_by_name(self):
        for name, ticker, addr, net in TEST_DATA:
            with self.subTest(name=name, address=addr, net=net):
                res = coinaddr.validate(name, addr)
                self.assertEqual(name, res.name)
                self.assertEqual(ticker, res.ticker)
                self.assertEqual(addr, res.address)
                self.assertEqual(True, res.valid)
                self.assertEqual(net, res.network)

    def test_validation_by_ticker(self):
        for name, ticker, addr, net in TEST_DATA:
            with self.subTest(name=name, ticker=ticker, address=addr, net=net):
                res = coinaddr.validate(ticker, addr)
                self.assertEqual(name, res.name)
                self.assertEqual(ticker, res.ticker)
                self.assertEqual(addr, res.address)
                self.assertEqual(True, res.valid)
                self.assertEqual(net, res.network)
                del res

    def test_validation_from_text(self):
        for name, ticker, addr, net in TEST_DATA:
            with self.subTest(name=name, address=addr, net=net):
                res = coinaddr.validate(name, addr.decode())
                self.assertEqual(name, res.name)
                self.assertEqual(ticker, res.ticker)
                self.assertEqual(addr, res.address)
                self.assertEqual(True, res.valid)
                self.assertEqual(net, res.network)


class TestExtendingCoinaddr(unittest.TestCase):
    def test_extending_currency(self):
        new_currency = Currency(
            'testcoin', ticker='ttc', validator='Base58Check',
            networks=dict(
                main=(0x00, 0x05), test=(0x6f, 0xc4)))

        self.assertEqual(new_currency, Currencies.get(new_currency.name))
        self.assertEqual(new_currency, Currencies.get(new_currency.ticker))

        test_data = [
            (new_currency.name, new_currency.ticker,
             b'1BoatSLRHtKNngkdXEeobR76b53LETtpyT', 'main')
            ]
        for name, ticker, addr, net in test_data:
            with self.subTest(name=name, ticker=ticker, address=addr, net=net):
                res = coinaddr.validate(name, addr)
                self.assertEqual(name, res.name)
                self.assertEqual(ticker, res.ticker)
                self.assertEqual(addr, res.address)
                self.assertEqual(True, res.valid)
                self.assertEqual(net, res.network)

            with self.subTest(name=name, ticker=ticker, address=addr, net=net):
                res = coinaddr.validate(ticker, addr)
                self.assertEqual(name, res.name)
                self.assertEqual(ticker, res.ticker)
                self.assertEqual(addr, res.address)
                self.assertEqual(True, res.valid)
                self.assertEqual(net, res.network)

    def test_extending_validator(self):
        class NewValidator(ValidatorBase):
            name = 'new'
            networks = 'testing'

            def validate(self):
                return True

        validator = Validators.get('new')
        self.assertEqual(NewValidator, validator)


if __name__ == '__main__':
    unittest.main()



"""
:mod:`coinaddr.currency`
~~~~~~~~~~~~~~~~~~~~~~~~

Containers for holding all the necessary data for validating cryptocurrencies.
"""

import attr
from zope.interface import implementer, provider

from .interfaces import ICurrency, INamedInstanceContainer
from .base import NamedInstanceContainerBase


@provider(INamedInstanceContainer)
class Currencies(metaclass=NamedInstanceContainerBase):
    """Container for all currencies."""

    @classmethod
    def get(cls, name, default=None):
        """Return currency object with matching name or ticker."""
        for inst in cls.instances.values():
            if name in (inst.name, inst.ticker):
                return inst
        else:
            return default


class CurrencyMeta(type):
    """Register currency classes on Currencies.currencies."""

    def __call__(cls, *args, **kwargs):
        inst = super(CurrencyMeta, cls).__call__(*args, **kwargs)
        Currencies[inst.name] = inst
        return inst


@implementer(ICurrency)
@attr.s(frozen=True, slots=True, cmp=False)
class Currency(metaclass=CurrencyMeta):
    """An immutable representation of a cryptocurrency specification."""

    name = attr.ib(
        type=str,
        validator=attr.validators.instance_of(str))
    ticker = attr.ib(
        type=str,
        validator=attr.validators.instance_of(str))
    validator = attr.ib(
        type='str',
        validator=attr.validators.instance_of(str))
    networks = attr.ib(
        type=dict,
        validator=attr.validators.optional(attr.validators.instance_of(dict)),
        default=attr.Factory(dict))
    charset = attr.ib(
        type=bytes,
        validator=attr.validators.optional(attr.validators.instance_of(bytes)),
        default=None)


Currency('bitcoin', ticker='btc', validator='Base58Check',
         networks=dict(
             main=(0x00, 0x05), test=(0x6f, 0xc4)))
Currency('bitcoin-cash', ticker='bch', validator='Base58Check',
         networks=dict(
             main=(0x00, 0x05), test=(0x6f, 0xc4)))
Currency('litecoin', ticker='ltc', validator='Base58Check',
         networks=dict(
             main=(0x30, 0x05, 0x32), test=(0x6f, 0xc4)))
Currency('dogecoin', ticker='doge', validator='Base58Check',
         networks=dict(
             main=(0x1e, 0x16), test=(0x71, 0xc4)))
Currency('dashcoin', ticker='dash', validator='Base58Check',
         networks=dict(
             main=(0x4c, 0x10), test=(0x8c, 0x13)))
Currency('neocoin', ticker='neo', validator='Base58Check',
         networks=dict(both=(0x17,)))
Currency('ripple', ticker='xrp', validator='Base58Check',
         networks=dict(both=(0x00, 0x05)),
         charset=(b'rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcd'
                  b'eCg65jkm8oFqi1tuvAxyz'))
Currency('ethereum', ticker='eth', validator='Ethereum')
Currency('ether-zero', ticker='etz', validator='Ethereum')
Currency('ethereum-classic', ticker='etc', validator='Ethereum')

Currency('zcash', ticker='zec', validator='Base58Check', 
         networks=dict(
             main=(0x1cb8, 0x1cbd), test=(0x1d25, 0x1cba)))



import unittest

from coinaddr.interfaces import (
    INamedSubclassContainer, IValidator, IValidationRequest, IValidationResult
    )
from coinaddr.validation import (
    Validators, ValidatorBase, ValidationRequest, ValidationResult,
    Base58CheckValidator, EthereumValidator
    )


class TestValidation(unittest.TestCase):
    def test_interfaces(self):
        self.assertTrue(INamedSubclassContainer.providedBy(Validators))

        validators = [Base58CheckValidator, EthereumValidator]
        for validator in validators:
            with self.subTest(validator=validator):
                self.assertTrue(IValidator.implementedBy(validator))

        self.assertTrue(
            IValidationRequest.implementedBy(ValidationRequest))
        self.assertTrue(
            IValidationResult.implementedBy(ValidationResult))


if __name__ == '__main__':
    unittest.main()



import unittest

from coinaddr.interfaces import INamedInstanceContainer, ICurrency
from coinaddr.currency import Currencies, Currency


class TestCurrency(unittest.TestCase):
    def test_interfaces(self):
        self.assertTrue(INamedInstanceContainer.providedBy(Currencies))
        self.assertTrue(ICurrency.implementedBy(Currency))

        for currency in Currencies.instances.values():
            with self.subTest(currency=currency):
                self.assertTrue(ICurrency.providedBy(currency))


if __name__ == '__main__':
    unittest.main()



import re
from setuptools import setup, find_packages


with open('coinaddrng/__init__.py', 'rt') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

try:
    from m2r import parse_from_file
    long_description = parse_from_file('README.md')
except ImportError:
    with open('README.md') as fd:
        long_description = fd.read()


setup(
    name='coinaddrng',
    version=version,
    description='A crypto-currency address inspection/validation library.',
    #long_description=long_description,
    keywords=[
        'bitcoin',
        'litecoin',
        'altcoin',
        'ethereum',
        'address',
        'validation',
        'inspection',
    ],
    author='Joe Black',
    author_email='me@joeblack.nyc',
    maintainer='Joe Black',
    maintainer_email='me@joeblack.nyc',
    url='https://github.com/joeblackwaslike/coinaddr',
    download_url=(
        'https://github.com/joeblackwaslike/coinaddr/tarball/v%s' % version),
    license='MIT',
    install_requires=[
        'attrs>=17.4.0',
        'pysha3>=1.0.2',
        'base58check>=1.0.1',
        'zope.interface>=4.4.3'
    ],
    zip_safe=False,
    packages=find_packages(),
    package_data={'': ['LICENSE']},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development',
        'Topic :: Security :: Cryptography',
        'Topic :: Text Processing',
        'Topic :: Utilities',
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)



# pylint: disable=inherit-non-class,no-self-argument,no-method-argument
# pylint: disable=unexpected-special-method-signature,arguments-differ

"""
:mod:`coinaddr.interfaces`
~~~~~~~~~~~~~~~~~~~~~~~~

Various zope compatible interfaces for the coinaddr package.
"""

from zope.interface import Interface, Attribute


class INamedInstanceContainer(Interface):
    """Contains all currencies instantiated."""

    instances = Attribute('Mapping of instance.name -> instance')

    def __getitem__(name):
        """Return the named instance"""

    def __setitem__(name, obj):
        """Add the named instance to the mapping of instances"""

    def __delitem__(name, obj):
        """Add the named instance to the mapping of instances"""

    def __contains__(name):
        """Return true if we contain the named instance"""

    def __iter__():
        """Return an iterable, iterating all instances"""

    def get(name, default=None):
        """Return the named instance if we contain it, else default"""


class INamedSubclassContainer(Interface):
    """Contains a weakvaluedict of subclasses."""

    subclasses = Attribute('Mapping of subclass.name -> subclass')

    def __getitem__(name):
        """Return the named subclass"""

    def __setitem__(name, obj):
        """Add the named subclass to the mapping of subclasses"""

    def __delitem__(name, obj):
        """Add the named subclass to the mapping of subclasses"""

    def __contains__(name):
        """Return true if we contain the named subclass"""

    def __iter__():
        """Return an iterable, iterating all subclasses"""

    def get(name, default=None):
        """Return the named subclass if we contain it, else default"""


class ICurrency(Interface):
    """A cryptocurrency address specification."""

    name = Attribute('Name of currency')
    ticker = Attribute('Ticker symbol for currency')
    validator = Attribute('Validator name for validation')
    networks = Attribute('The networks and version bytes for those networks')
    charset = Attribute('For base58Check based currencies, custom charset.')


class IValidator(Interface):
    """A cryptocurrency address validator."""

    name = Attribute('Name of validator')
    network = Attribute('Network name of address being validated')

    def validate():
        """Validate the address type, True if valid, else False."""


class IValidationRequest(Interface):
    """Contains the data and helpers for a given validation request."""

    currency = Attribute('The currency name or ticker being validated')
    address = Attribute('The address to be validated')
    extras = Attribute('Any extra attributes to be passed to decoder, etc')
    networks = Attribute(
        'Concatenated list of all network versions for currency')

    def execute():
        """Executes the request and returns a ValidationResult object"""


class IValidationResult(Interface):
    """Represents all data for a validation result."""

    name = Attribute('Name of currency for address validated')
    ticker = Attribute('Ticker of currency for address validated')
    address = Attribute('The address that was validated')
    valid = Attribute('Boolean representing whether the address is valid')
    network = Attribute(
        'Name of network the address belongs to if applicable')



"""
:mod:`coinaddr.validation`
~~~~~~~~~~~~~~~~~~~~~~~~

Base/meta-classes for the currency and validation machinery.
"""

from weakref import WeakValueDictionary


class NamedInstanceContainerBase(type):
    """A Container for instances."""

    def __init__(cls, name, bases, idict):
        super(NamedInstanceContainerBase, cls).__init__(name, bases, idict)
        cls.instances = dict()

    def __getitem__(cls, name):
        return cls.instances[name]

    def __setitem__(cls, name, obj):
        cls.instances[name] = obj

    def __delitem__(cls, name):
        del cls.instances[name]

    def __contains__(cls, name):
        return name in cls.instances

    def __iter__(cls):
        return iter(cls.instances)

    def get(cls, name, default=None):
        """Returns instance by name."""
        return cls.instances.get(name, default)


class NamedSubclassContainerBase(type):
    """A Container for subclasses."""

    def __init__(cls, name, bases, idict):
        super(NamedSubclassContainerBase, cls).__init__(name, bases, idict)
        cls.subclasses = WeakValueDictionary()

    def __getitem__(cls, name):
        return cls.subclasses[name]

    def __setitem__(cls, name, obj):
        cls.subclasses[name] = obj

    def __delitem__(cls, name):
        del cls.subclasses[name]

    def __contains__(cls, name):
        return name in cls.subclasses

    def __iter__(cls):
        return iter(cls.subclasses)

    def get(cls, name, default=None):
        """Returns subclass by name."""
        return cls.subclasses.get(name, default)
